import textwrap

from .spm_object import SPMObject

class ImageCalculator(SPMObject):
    """ The image calculator is for performing user-specified algebraic 
        manipulations on a set of images. The result is being written out as an 
        image. The user is prompted to supply images to work on, a filename for 
        the output  image, and the expression to evaluate. The expression should 
        be a standard MATLAB expression, within which the images should be 
        referred to as i1, i2, i3,... etc.
    """
    
    def __init__(self, inputs, output, expression, dtype="int16"):
        SPMObject.__init__(self, "spm.util.imcalc")
        
        self.inputs = inputs
        self.output = output
        self.expression = expression
        self.dtype = dtype
        
        self._dtypes_map = {
            "uint8": 2,
            "int16": 4,
            "int32": 8,
            "float32": 16,
            "float64": 64,
            "int8": 256,
            "uint16": 512,
            "uint32": 768,
        }
        
        self.template = textwrap.dedent("""\
            {{ id(index, name) }}.input = {
            {% for input in inputs -%}
            {{ ((id(index, name)+".input = {")|length)*" " }}'{{ input }}'
            {% endfor -%}
            {{ ((id(index, name)+".input = {")|length)*" " }}};
            {{ id(index, name) }}.output = '{{ output.name }}';
            {{ id(index, name) }}.outdir = {'{{ output.parent }}'};
            {{ id(index, name) }}.expression = '{{ expression }}';
            {{ id(index, name) }}.var = struct('name', {}, 'value', {});
            {{ id(index, name) }}.options.dmtx = 0;
            {{ id(index, name) }}.options.mask = 0;
            {{ id(index, name) }}.options.interp = 1;
            {{ id(index, name) }}.options.dtype = {{ _dtypes_map[dtype] }};""")
    
    def _get_file_dep(self):
        return sorted(set(seilf.inputs))
    
    def _get_targets(self):
        return [self.output]



import itertools
import pathlib
import textwrap

from .spm_object import SPMObject

class Masking(SPMObject):
    """ The mask specifies the voxels within the image volume which are to be
        assessed. SPM supports three methods of masking: (1) Threshold, 
        (2) Implicit and (3) Explicit. The volume analysed is the intersection
        of all masks.
    """
    
    def __init__(self, 
            threshold_mode=None, threshold=None, implicit=True, explicit=""):
        
        SPMObject.__init__(self, "spm.stats.factorial_design.masking")
        if threshold_mode not in [None, "absolute", "relative"]:
            raise Exception("Invalid threshold mode: {}".format(threshold_mode))
        self.threshold_mode = threshold_mode
        
        if threshold_mode == "absolute" and threshold is None:
            self.threshold = 100
        elif threshold_mode == "relative" and threshold is None:
            self.threshold = 0.8
        else:
            self.threshold = threshold
        
        self.implicit = implicit
        self.explicit = explicit
        
        self.template = textwrap.dedent("""\
            {%- if threshold_mode == none -%}
            {{ id(index, name) }}.tm.tm_none = 1;
            {%- elif threshold_mode == "absolute" -%}
            {{ id(index, name) }}.tma.athresh = {{ threshold }};
            {%- elif threshold_mode == "relative" -%}
            {{ id(index, name) }}.tmr.rthresh = {{ threshold }};
            {%- endif %}
            {{ id(index, name) }}.im = {{ implicit|int }};
            {{ id(index, name) }}.em = {'{{ explicit}}'};""")
    
    def _get_file_dep(self):
        return [self.explicit] if self.explicit else []

class GlobalCalculation(SPMObject):
    """ This option is for PET or VBM data (not second level fMRI).
        
        There  are three methods for estimating global effects (1) Omit 
        (assumming no other options requiring the global value chosen) (2) User 
        defined (enter your own vector of global values) (3) Mean: SPM standard 
        mean voxel value (within per image fullmean/8 mask).
    """
    
    def __init__(self, mode="omit", values=None):
        SPMObject.__init__(self, "spm.stats.factorial_design.globalc")
        
        if mode not in ["omit", "user", "mean"]:
            raise Exception("Invalid mode: {}".format(mode))
        self.mode = mode
        
        self.values = values
        
        self.template = textwrap.dedent("""\
            {%- if mode == "omit" -%}
            {{ id(index, name) }}.g_omit = 1;
            {%- elif mode == "user" -%}
            {{ id(index, name) }}.g_user.global_uval = [{{ values|join(";") }}];
            {%- elif mode == "mean" -%}
            {{ id(index, name) }}.g_mean = 1;
            {%- endif -%}""")

class GlobalNormalization(SPMObject):
    """ These options are for PET or VBM data (not second level fMRI).
        
        'Overall grand mean scaling'  imply scales all the data by a common 
        factor such that the mean of all the global values is the value 
        specified.
        
        'Normalisation' refers to either proportionally scaling each image or 
        adding a covariate to adjust for the global values.
    """
    
    def __init__(self, value=None, mode=None):
        SPMObject.__init__(self, "spm.stats.factorial_design.globalm")
        
        self.value = value
        
        if mode not in [None, "proportional", "ancova"]:
            raise Exception("Invalid mode: {}".format(mode))
        self.mode = {None: 1, "proportional": 2, "ancova": 3}[mode]
        
        self.template = textwrap.dedent("""\
            {%- if value == none -%}
            {{ id(index, name) }}.gmsca.gmsca_no = 1;
            {%- else -%}
            {{ id(index, name) }}.gmsca.gmsca_yes.gmscv = {{ value }};
            {%- endif %}
            {{ id(index, name) }}.glonorm = {{ mode }};""")

class Covariate(object):
    """ Covariate of the experimental design.
    """
    
    def __init__(self, cname, values, interactions=None, centering="Overall mean"):
        self.cname = cname
        self.values = values
        
        known_interactions = [None, "With Factor 1", "With Factor 2", "With Factor 3"]
        if interactions not in known_interactions:
            raise Exception("Invalid interactions: {}".format(interactions))
        self.interactions = 1+known_interactions.index(interactions)
        
        known_centering = [
                "Overall mean", 
                "Factor 1 mean", "Factor 2 mean", "Factor 3 mean",
                "No centering", "User specified value", "As implied by ANCOVA", 
                "GM"]
        
        if centering not in known_centering:
            raise Exception("Invalid centering: {}".format(centering))
        self.centering = 1+known_centering.index(centering)

class Covariates(SPMObject):
    """ This  option  allows  for the specification of covariates and nuisance 
        variables (note that SPM does not make any distinction between effects 
        of interest (including covariates) and nuisance effects).
    """
    
    def __init__(self, covariates):
        SPMObject.__init__(self, "spm.stats.factorial_design.cov")
        self.covariates = covariates
        self.template = textwrap.dedent("""\
            {%- if covariates -%}
            {% for covariate in covariates -%}
            {{ id(index, name) }}({{ loop.index }}).c = [
            {% for value in covariate.values -%}
            {{ ((id(index, name)+"("+(loop.index|string)+").c = {")|length)*" " }}{{ value }}
            {% endfor -%}
            {{ ((id(index, name)+"("+(loop.index|string)+").c = {")|length)*" " }}];
            {{ id(index, name) }}({{ loop.index }}).cname = '{{ covariate.cname }}';
            {{ id(index, name) }}({{ loop.index }}).iCFI = {{ covariate.interactions }};
            {{ id(index, name) }}({{ loop.index }}).iCC = {{ covariate.centering }};
            {% endfor -%}
            {%- else -%}
            {{ id(index, name) }} = struct('c', {}, 'cname', {}, 'iCFI', {}, 'iCC', {});
            {%- endif -%}""")

# Mutliple covariates

class OneSampleTTest(SPMObject):
    def __init__(self, scans):
        SPMObject.__init__(self, "spm.stats.factorial_design.des.t1")
        self.scans = scans
        self.template = textwrap.dedent("""\
            {{ id(index, name) }}.scans = {
            {% for scan in scans -%}
            {{ ((id(index, name)+".scans = {")|length)*" " }}'{{ scan }}'
            {% endfor -%}
            {{ ((id(index, name)+".scans = {")|length)*" " }}};
            """)
    
    def _get_file_dep(self):
        return sorted(set(self.scans))

class TwoSamplesTTest(SPMObject):
    def __init__(
            self, scans1, scans2, 
            independence=True, equal_variance=False, grand_mean_scaling=False,
            ancova=False):
        
        SPMObject.__init__(self, "spm.stats.factorial_design.des.t2")
        self.scans1 = scans1
        self.scans2 = scans2
        self.independence = independence
        self.equal_variance = equal_variance
        self.grand_mean_scaling = grand_mean_scaling
        self.ancova = ancova
        
        self.template = textwrap.dedent("""\
            {{ id(index, name) }}.scans1 = {
            {% for scan in scans1 -%}
            {{ ((id(index, name)+".scans1 = {")|length)*" " }}'{{ scan }}'
            {% endfor -%}
            {{ ((id(index, name)+".scans1 = {")|length)*" " }}};
            {{ id(index, name) }}.scans2 = {
            {% for scan in scans2 -%}
            {{ ((id(index, name)+".scans1 = {")|length)*" " }}'{{ scan }}'
            {% endfor -%}
            {{ ((id(index, name)+".scans1 = {")|length)*" " }}};
            {{ id(index, name) }}.dept = {{ (not independence)|int }};
            {{ id(index, name) }}.variance = {{ (not equal_variance)|int }};
            {{ id(index, name) }}.gmsca = {{ grand_mean_scaling|int }};
            {{ id(index, name) }}.ancova = {{ ancova|int }};
            """)
    
    def _get_file_dep(self):
        return sorted(set(itertools.chain(self.scans1, self.scans2)))

class PairedTTest(SPMObject):
    def __init__(self, pairs, grand_mean_scaling=False, ancova=False):
        
        SPMObject.__init__(self, "spm.stats.factorial_design.des.pt")
        self.pairs = pairs
        self.grand_mean_scaling = grand_mean_scaling
        self.ancova = ancova
        
        self.template = textwrap.dedent("""\
            {% for scan1, scan2 in pairs -%}
            {{ id(index, name) }}.pair({{ loop.index }}) = {
            {{ ((id(index, name)+".pair("+(loop.index|string)+") = {")|length)*" " }}'{{ scan1 }}'
            {{ ((id(index, name)+".pair("+(loop.index|string)+") = {")|length)*" " }}'{{ scan2 }}'
            {{ ((id(index, name)+".pair("+(loop.index|string)+") = {")|length)*" " }}};
            {% endfor -%}
            
            {{ id(index, name) }}.gmsca = {{ grand_mean_scaling|int }};
            {{ id(index, name) }}.ancova = {{ ancova|int }};
            """)
    
    def _get_file_dep(self):
        return sorted(set(itertools.chain(*self.pairs)))

class ANOVA(SPMObject):
    """ One-way Analysis of Variance (ANOVA).
    """
    
    def __init__(
            self, cells,
            independance=True, equal_variance=False, grand_mean_scaling=False,
            ancova=False):
        
        SPMObject.__init__(self, "spm.stats.factorial_design.des.anova")
        self.cells = cells
        self.independance = independance
        self.equal_variance = equal_variance
        self.grand_mean_scaling = grand_mean_scaling
        self.ancova = ancova
        
        self.template = textwrap.dedent("""\
            {% for cell in cells -%}
            {{ id(index, name) }}.icell({{ loop.index }}).scans = {
            {% set padding=((id(index, name)+".icell("+(loop.index|string)+").scans = {")|length)*" " -%}
            {% for scan in cell -%}
            {{ padding }}'{{ scan }}'
            {% endfor -%}
            {{ padding }}};
            {% endfor -%}
            {{ id(index, name) }}.dept = {{ (not independance)|int }};
            {{ id(index, name) }}.variance = {{ (not equal_variance)|int }};
            {{ id(index, name) }}.gmsca = {{ grand_mean_scaling|int }};
            {{ id(index, name) }}.ancova = {{ ancova|int }};
            """)
    
    def _get_file_dep(self):
        return sorted(set(itertools.chain(*self.cells)))

class FactorialDesign(SPMObject):
    """ Configuration  of the design matrix, describing the general linear 
        model, data specification, and other parameters necessary for the 
        statistical analysis.
        These  parameters are saved in a configuration file (SPM.mat), which can
        then be passed on to ModelEstimation which estimates the design. 
        Inference on these estimated parameters is then handled by the SPM 
        results section.
    """
    
    def __init__(
            self, output_directory, design, covariates=None,
            masking=None, global_calculation=None, global_normalization=None):
        
        SPMObject.__init__(self, "spm.stats.factorial_design")
        
        self.output_directory = output_directory
        self.design = design
        self.covariates = covariates or Covariates([])
        self.masking = masking or Masking()
        self.global_calculation = global_calculation or GlobalCalculation()
        self.global_normalization = global_normalization or GlobalNormalization()
    
        self.template = textwrap.dedent("""\
            {{ id(index, name) }}.dir = {'{{ output_directory }}'};
            {{ _design }}
            {{ _covariates }}
            {{ id(index, name) }}.multi_cov = struct('files', {}, 'iCFI', {}, 'iCC', {});
            {{ _masking }}
            {{ _global_calculation }}
            {{ _global_normalization }}
            """)
    
    @property
    def spmmat(self):
        return pathlib.Path(self.output_directory)/"SPM.mat"
    
    def get_script(self, index):
        self._design = self.design.get_script(index)
        self._covariates = self.covariates.get_script(index)
        self._masking = self.masking.get_script(index)
        self._global_calculation = self.global_calculation.get_script(index)
        self._global_normalization = self.global_normalization.get_script(index)
        return SPMObject.get_script(self, index)
    
    def _get_file_dep(self):
        return list(itertools.chain(self.design.file_dep, self.masking.file_dep))
    
    def _get_targets(self):
        return [self.spmmat]



import textwrap

from .spm_object import SPMObject

class ModelEstimation(SPMObject):
    """ Estimation of model parameters using classical (ReML - Restricted 
        Maximum Likelihood) or Bayesian algorithms. After  parameter estimation,
        the 'Results' button can be used to specify contrasts that will produce 
        Statistical Parametric Maps (SPMs) or Posterior Probability Maps (PPMs) 
        and tables of statistics.
    """
    
    def __init__(self, design, write_residuals=False, method="Classical"):
        SPMObject.__init__(self, "spm.stats.fmri_est")
        
        self.design = design
        self.write_residuals = write_residuals
        
        if method not in ["Classical", "Bayesian2"]:
            raise Exception("Unknown method: {}".format(method))
        self.method = method
        
        self.template = textwrap.dedent("""\
            {{ id(index, name) }}.spmmat = {'{{ design.spmmat }}'};
            {{ id(index, name) }}.write_residuals = {{ write_residuals|int }};
            {{ id(index, name) }}.method.{{ method }} = 1;""")
    
    def _get_file_dep(self):
        return [self.design.spmmat]
    
    def _get_targets(self):
        directory = self.design.spmmat.parent
        targets = [
            self.design.spmmat, 
            directory/"mask.nii", directory/"ResMS.nii", directory/"RPV.nii"]
        
        # FIXME: the parameter maps ("beta_XXXX.nii") are missing from the list.
        # How can we compute the from self.design.design?
            
        return targets



import inspect

import doit

class TaskGraph(object):
    def __init__(self):
        self.tasks = {}
        self.children = {}
    
    def build(self, creators):
        self.tasks = self._get_tasks(creators)
        
        for name, (doit_task, spire_task) in self.tasks.items():
            dependencies = doit_task.setup_tasks + doit_task.task_dep
            for dependency in dependencies:
                self.children.setdefault(dependency, []).append(doit_task.name)
    
    def _get_tasks(self, creators):
        doit_tasks = []
        spire_tasks = {}
        for name, creator, line in creators:
            result = creator()
            
            if isinstance(result, dict):
                dicts = [result]
            elif inspect.isgenerator(result):
                dicts = [x for x in result]
            elif result is None:
                continue
            
            for dict_ in dicts:
                dict_["name"] = dict_.pop("basename", creator.__name__)
                # Replace None with "" to avoid an error on task creation
                dict_["file_dep"] = [x or "" for x in dict_["file_dep"]]
                
                doit_tasks.append(doit.task.dict_to_task(dict_))
                spire_tasks[dict_["name"]] = creator.__self__
        
        doit_tasks = doit.control.TaskControl(doit_tasks).tasks
        if sorted(list(doit_tasks.keys())) != sorted(list(spire_tasks.keys())):
            raise Exception("doit tasks and Spire tasks do not match")
        
        tasks = {
            key: (doit_tasks[key], spire_tasks[key]) 
            for key in doit_tasks.keys()}
        
        return tasks



import os
import subprocess
import tempfile

def find(matlab="matlab", matlab_path=None):
    """ Return the root directory of SPM. matlab, if given, is the path to the
        MATLAB executable. matlab_path, if given, is a MATLAB expression fed to
        addpath.
    """
    
    script = ("if isempty(which('spm'));"
              "fprintf(1, '');"
              "exit();"
              "end;"
              "fprintf(1, '%s', spm('dir'));"
              "exit();")

    if matlab_path:
        script = "addpath({});".format(matlab_path)+script

    try:
        output = subprocess.check_output([
            matlab, "-nodisplay", "-nosplash", "-nojvm", "-r", script])
    except subprocess.CalledProcessError as e:
        raise Exception("Could not find SPM: {}".format(e))
    last_line = output.splitlines()[-1]
    # Weird data at the end of the line
    if b"\x1b" in last_line :
        last_line = last_line[:last_line.index(b"\x1b")]
    if last_line == "" :
        raise Exception("Could not find SPM")
    
    return last_line.decode()

def get_script(tools, standalone=True, exit=True, modality="fmri"):
    """ Return a Matlab script running the provided tools. If standalone is 
        True, then the script will include SPM initialization commands and the
        resulting script can be directly fed to Matlab. If standalone is False,
        the resulting script can be loaded in SPM Batch Editor.
    """
    
    script = []
    
    if standalone:
        script.extend([
            "spm('defaults','{}');".format(modality),
            "spm_jobman('initcfg');"
        ])
    
    for index, tool in enumerate(tools):
        script.append(tool.get_script(1+index))
    
    if standalone:
        script.append("spm_jobman('run',matlabbatch);")
        if exit:
            script.append("exit();")
    
    return "\n".join(script)

def run(jobs, matlab="matlab"):
    """ Run a list of jobs in Matlab. The ``matlab`` argument, if specified,
        is the path to the MATLAB executable.
    """
    
    fd, path = tempfile.mkstemp(suffix=".m")
    os.write(fd, "cd {}; {}".format(os.getcwd(), get_script(jobs)).encode())
    os.close(fd)

    try:
        subprocess.check_call([
            matlab, "-nodisplay", "-nosplash", "-r", "run('{}');".format(path)])
    finally:
        os.remove(path)



import inspect
import itertools
import logging

import doit

from .task import Task
from .task_factory import TaskFactory
from .task_graph import TaskGraph

################################################################################
# Replace doit's task loader to include objects created by TaskFactory         #
################################################################################

doit_loader_load_tasks = None
def spire_load_tasks(*args, **kwargs):
    tasks = doit_loader_load_tasks(*args, **kwargs)
    
    for object_ in TaskFactory._task_registry:
        if getattr(object_, "skipped", False):
            continue
        
        dict_ = object_.as_task_dict()
        if dict_ is None:
            continue
        
        dict_["name"] = dict_.pop("basename")
        # Replace None with "" to avoid an error on task creation
        dict_["file_dep"] = [x or "" for x in dict_["file_dep"]]
            
        tasks.append(doit.task.dict_to_task(dict_))
            
    return tasks
    
if doit.loader.load_tasks != spire_load_tasks:
    doit_loader_load_tasks = doit.loader.load_tasks
    doit.loader.load_tasks = spire_load_tasks

################################################################################
# Prune the task graph from tasks having None in their file_dep                #
################################################################################

def prune():
    task_list = []
    spire_task_dict = {}
    
    # Look at creators in the caller's globals 
    caller = inspect.getouterframes(inspect.currentframe())[1][0]
    
    # https://github.com/pydoit/doit/blob/0.31.1/doit/loader.py#L129-L133
    creators = doit.loader._get_task_creators(caller.f_globals, [])
    creators.sort(key=lambda obj: obj[2])
    creators.extend(
        [(None, x.as_task_dict, None) for x in TaskFactory._task_registry])
    
    graph = TaskGraph()
    graph.build(creators)
    
    to_skip = []
    for (doit_task, spire_task) in graph.tasks.values():
        if "" in doit_task.file_dep:
            to_skip.append(doit_task.name)
    
    # Breadth-first traversal of the dependency tree, marking request tasks and
    # their descendents as skipped
    while len(to_skip) > 0:
        root = to_skip.pop()
        doit_task, spire_task = graph.tasks[root]
        if not getattr(spire_task, "skipped", False):
            logging.warning("Skipping {}".format(root))
            spire_task.skipped = True
            to_skip.extend(graph.children.get(root, []))

################################################################################
# Representation of the task graph in the Graphviz format                      #
################################################################################

def graph(tasks_only=False):
    task_list = []
    spire_task_dict = {}
    
    # Look at creators in the caller's globals 
    caller = inspect.getouterframes(inspect.currentframe())[1][0]
    
    # https://github.com/pydoit/doit/blob/0.31.1/doit/loader.py#L129-L133
    creators = doit.loader._get_task_creators(caller.f_globals, [])
    creators.sort(key=lambda obj: obj[2])
    creators.extend(
        [(None, x.as_task_dict, None) for x in TaskFactory._task_registry])
    
    graph = TaskGraph()
    graph.build(creators)
    
    def quote(name):
        return "\"{}\"".format(name.replace("\"", "\\\""))
    
    lines = ["digraph {"]
    
    deps_and_targets = set()
    for name, (doit_task, _) in graph.tasks.items():
        lines.append("    {}[shape=box];".format(quote(name)))
        if not tasks_only:
            for entry in itertools.chain(doit_task.file_dep, doit_task.targets):
                if entry not in deps_and_targets:
                    deps_and_targets.add(entry)
                    lines.append(
                        "    {}[shape=parallelogram];".format(quote(entry)))
        
            # We can create the edges now: the task, deps and targets node are
            # already present
            for entry in doit_task.file_dep:
                lines.append("    {} -> {};".format(quote(entry), quote(name)))
            for entry in doit_task.targets:
                lines.append("    {} -> {};".format(quote(name), quote(entry)))
    
    if tasks_only:
        # We need to wait for all task nodes to be created
        for name, children in graph.children.items():
            lines.extend(
                "    {} -> {};".format(quote(name), quote(x)) 
                for x in children)
            

    lines.extend(["}", ""])
    
    return "\n".join(lines)
